@@ -13,7 +13,7 @@ from matplotlib import style
 from seldonian.utils.io_utils import save_pickle
 
 from .experiments import BaselineExperiment, SeldonianExperiment, FairlearnExperiment
-from .utils import generate_resampled_datasets
+from .experiment_utils import generate_resampled_datasets, has_failed
 
 seldonian_model_set = set(["qsa","headless_qsa", "sa"])
 plot_colormap = matplotlib.cm.get_cmap("tab10")
@@ -142,23 +142,17 @@ class PlotGenerator:
         :type savename: str, defaults to None
         """
         plt.style.use("bmh")
-        # plt.style.use('grayscale')
         regime = self.spec.dataset.regime
-        tot_data_size = self.spec.dataset.num_datapoints
+        if regime == "supervised_learning":
+            tot_data_size = self.spec.dataset.num_datapoints
+        else:
+            tot_data_size = self.hyperparameter_and_setting_dict['num_episodes']
 
         # Read in constraints
         parse_trees = self.spec.parse_trees
-
-        constraint_dict = {}
-        for pt_ii, pt in enumerate(parse_trees):
-            delta = pt.delta
-            constraint_str = pt.constraint_str
-            constraint_dict[f"constraint_{pt_ii}"] = {
-                "delta": delta,
-                "constraint_str": constraint_str,
-            }
-
-        constraints = list(constraint_dict.keys())
+        n_constraints = len(parse_trees)
+        constraint_strs = [pt.constraint_str for pt in parse_trees]
+        deltas = [pt.delta for pt in parse_trees]
 
         # Figure out what experiments we have from subfolders in results_dir
         subfolders = [
@@ -184,6 +178,12 @@ class PlotGenerator:
             df_baseline["solution_returned"] = df_baseline["performance"].apply(
                 lambda x: ~np.isnan(x)
             )
+            df_baseline['gvec'] = df_baseline['gvec'].apply(lambda t: np.fromstring(t[1:-1],sep=' '))
+            new_colnames = ['g' + str(ii) + '_failed' for ii in range(1,n_constraints+1)]
+            for ii in range(len(new_colnames)):
+                colname = new_colnames[ii]
+                df_baseline[colname] = df_baseline['gvec'].str[ii].apply(has_failed)
+            df_baseline = df_baseline.drop('gvec', axis=1)
 
             valid_mask = ~np.isnan(df_baseline["performance"])
             df_baseline_valid = df_baseline[valid_mask]
@@ -210,6 +210,13 @@ class PlotGenerator:
             )
 
             df_seldonian = pd.read_csv(savename_seldonian)
+            df_seldonian['gvec'] = df_seldonian['gvec'].apply(lambda t: np.fromstring(t[1:-1],sep=' '))
+            new_colnames = ['g' + str(ii) + '_failed' for ii in range(1,n_constraints+1)]
+            for ii in range(len(new_colnames)):
+                colname = new_colnames[ii]
+                df_seldonian[colname] = df_seldonian['gvec'].str[ii].apply(has_failed)
+            df_seldonian = df_seldonian.drop('gvec', axis=1)
+
             passed_mask = df_seldonian["passed_safety"] == True
             df_seldonian_passed = df_seldonian[passed_mask]
             # Get the list of all data_fracs
@@ -226,22 +233,25 @@ class PlotGenerator:
             seldonian_dict[seldonian_model]["X_passed"] = X_passed
 
         ## PLOTTING SETUP
+        vert_size = 3 + n_constraints
         if include_legend:
-            figsize = (9, 4.5)
+            vert_size+=0.5
+            figsize = (9, vert_size)
         else:
-            figsize = (9, 4)
+            figsize = (9, vert_size)
         fig = plt.figure(figsize=figsize)
         plot_index = 1
-        n_rows = len(constraints)
+        n_rows = len(constraint_strs)
         n_cols = 3
         fontsize = fontsize
         legend_fontsize = legend_fontsize
         legend_handles = []
         legend_labels = []
-        ## Loop over constraints and make three plots for each constraint
-        for ii, constraint in enumerate(constraints):
-            constraint_str = constraint_dict[constraint]["constraint_str"]
-            delta = constraint_dict[constraint]["delta"]
+
+        # One row per constraint
+        for constraint_index, constraint_str in enumerate(constraint_strs):
+            constraint_num = constraint_index+1
+            delta = deltas[constraint_index]
 
             # SETUP FOR PLOTTING
             ax_performance = fig.add_subplot(n_rows, n_cols, plot_index)
@@ -261,14 +271,11 @@ class PlotGenerator:
 
             # Plot labels
             ax_performance.set_ylabel(performance_label, fontsize=fontsize)
-            ax_sr.set_ylabel("Probability of solution", fontsize=fontsize)
-            ax_fr.set_ylabel("Probability constraint was violated", fontsize=fontsize)
+            ax_sr.set_ylabel("Prob. of solution", fontsize=fontsize)
+            ax_fr.set_ylabel("Prob. of violation", fontsize=fontsize)
 
             # Only put horizontal axis labels on last row of plots
-            if ii == len(constraints) - 1:
-                # ax_performance.set_xlabel('Training samples',fontsize=fontsize)
-                # ax_sr.set_xlabel('Training samples',fontsize=fontsize)
-                # ax_fr.set_xlabel('Training samples',fontsize=fontsize)
+            if constraint_index == n_constraints - 1:
                 ax_performance.set_xlabel("Amount of data", fontsize=fontsize)
                 ax_sr.set_xlabel("Amount of data", fontsize=fontsize)
                 ax_fr.set_xlabel("Amount of data", fontsize=fontsize)
@@ -315,16 +322,17 @@ class PlotGenerator:
                 baseline_ste_performance = baseline_std_performance / np.sqrt(n_trials)
                 X_valid_baseline = this_baseline_dict["X_valid"]
                 (pl,) = ax_performance.plot(
-                    X_valid_baseline,
-                    baseline_mean_performance,
+                    X_valid_baseline.to_numpy(),
+                    baseline_mean_performance.to_numpy(),
                     color=baseline_color,
                     label=baseline,
                 )
-                legend_handles.append(pl)
-                if baseline in model_label_dict:
-                    legend_labels.append(model_label_dict[baseline])
-                else:
-                    legend_labels.append(baseline)
+                if constraint_index == 0:
+                    legend_handles.append(pl)
+                    if baseline in model_label_dict:
+                        legend_labels.append(model_label_dict[baseline])
+                    else:
+                        legend_labels.append(baseline)
                 ax_performance.scatter(
                     X_valid_baseline,
                     baseline_mean_performance,
@@ -340,32 +348,34 @@ class PlotGenerator:
                     alpha=0.5,
                 )
 
+            # Seldonian performance
             for seldonian_i, seldonian_model in enumerate(seldonian_models):
                 this_seldonian_dict = seldonian_dict[seldonian_model]
                 seldonian_color = plot_colormap(seldonian_i)
                 df_seldonian_passed = this_seldonian_dict["df_seldonian_passed"]
                 mean_performance = df_seldonian_passed.groupby("data_frac").mean()[
                     "performance"
-                ]
+                ].to_numpy()
                 std_performance = df_seldonian_passed.groupby("data_frac").std()[
                     "performance"
-                ]
+                ].to_numpy()
                 n_passed = df_seldonian_passed.groupby("data_frac").count()[
                     "performance"
-                ]
+                ].to_numpy()
                 ste_performance = std_performance / np.sqrt(n_passed)
-                X_passed_seldonian = this_seldonian_dict["X_passed"]
+                X_passed_seldonian = this_seldonian_dict["X_passed"].to_numpy()
                 (pl,) = ax_performance.plot(
                     X_passed_seldonian,
                     mean_performance,
                     color=seldonian_color,
                     linestyle="--",
                 )
-                legend_handles.append(pl)
-                if seldonian_model in model_label_dict:
-                    legend_labels.append(model_label_dict[seldonian_model])
-                else:
-                    legend_labels.append(seldonian_model)
+                if constraint_index == 0:
+                    legend_handles.append(pl)
+                    if seldonian_model in model_label_dict:
+                        legend_labels.append(model_label_dict[seldonian_model])
+                    else:
+                        legend_labels.append(seldonian_model)
 
                 ax_performance.scatter(
                     X_passed_seldonian,
@@ -393,15 +403,14 @@ class PlotGenerator:
             # to run model.fit() )
             for baseline_i, baseline in enumerate(baselines):
                 this_baseline_dict = baseline_dict[baseline]
-                X_all_baseline = this_baseline_dict["X_all"]
                 baseline_color = plot_colormap(baseline_i + len(seldonian_models))
                 df_baseline = this_baseline_dict["df_baseline"]
                 n_trials = df_baseline["trial_i"].max() + 1
-                mean_sr = df_baseline.groupby("data_frac").mean()["solution_returned"]
-                std_sr = df_baseline.groupby("data_frac").std()["solution_returned"]
+                mean_sr = df_baseline.groupby("data_frac").mean()["solution_returned"].to_numpy()
+                std_sr = df_baseline.groupby("data_frac").std()["solution_returned"].to_numpy()
                 ste_sr = std_sr / np.sqrt(n_trials)
 
-                X_all_baseline = this_baseline_dict["X_all"]
+                X_all_baseline = this_baseline_dict["X_all"].to_numpy()
 
                 ax_sr.plot(
                     X_all_baseline, mean_sr, color=baseline_color, label=baseline
@@ -421,16 +430,17 @@ class PlotGenerator:
                     alpha=0.5,
                 )
 
+            # Seldonian solution rate
             for seldonian_i, seldonian_model in enumerate(seldonian_models):
                 this_seldonian_dict = seldonian_dict[seldonian_model]
                 seldonian_color = plot_colormap(seldonian_i)
                 df_seldonian = this_seldonian_dict["df_seldonian"]
                 n_trials = df_seldonian["trial_i"].max() + 1
-                mean_sr = df_seldonian.groupby("data_frac").mean()["passed_safety"]
-                std_sr = df_seldonian.groupby("data_frac").std()["passed_safety"]
+                mean_sr = df_seldonian.groupby("data_frac").mean()["passed_safety"].to_numpy()
+                std_sr = df_seldonian.groupby("data_frac").std()["passed_safety"].to_numpy()
                 ste_sr = std_sr / np.sqrt(n_trials)
 
-                X_all_seldonian = this_seldonian_dict["X_all"]
+                X_all_seldonian = this_seldonian_dict["X_all"].to_numpy()
 
                 ax_sr.plot(
                     X_all_seldonian,
@@ -461,54 +471,68 @@ class PlotGenerator:
             ##########################
 
             # Baseline failure rate
-            
             for baseline_i, baseline in enumerate(baselines):
                 baseline_color = plot_colormap(baseline_i + len(seldonian_models))
                 # Baseline performance
                 this_baseline_dict = baseline_dict[baseline]
-                df_baseline_valid = this_baseline_dict["df_baseline_valid"]
-                n_trials = df_baseline_valid["trial_i"].max() + 1
+                df_baseline = this_baseline_dict["df_baseline"]
+                n_trials = df_baseline["trial_i"].max() + 1
 
-                baseline_mean_fr = df_baseline_valid.groupby("data_frac").mean()[
-                    "failed"
-                ]
-                baseline_std_fr = df_baseline_valid.groupby("data_frac").std()["failed"]
+                # baseline_mean_fr = df_baseline_valid.groupby("data_frac").mean()[
+                #     "failed"
+                # ].to_numpy()
+                # baseline_std_fr = df_baseline_valid.groupby("data_frac").std()["failed"].to_numpy()
+                gstr_failed = 'g' + str(constraint_num) + '_failed'
+
+                baseline_mean_fr = df_baseline.groupby("data_frac").mean()[
+                    gstr_failed].to_numpy()
+                # Need to groupby data frac
+                baseline_std_fr = df_baseline.groupby("data_frac").std()[
+                    gstr_failed].to_numpy()
                 baseline_ste_fr = baseline_std_fr / np.sqrt(n_trials)
 
-                X_valid_baseline = this_baseline_dict["X_valid"]
+                X_all_baseline = this_baseline_dict["X_all"].to_numpy()
 
                 ax_fr.plot(
-                    X_valid_baseline,
+                    X_all_baseline,
                     baseline_mean_fr,
                     color=baseline_color,
                     label=baseline,
                 )
                 ax_fr.scatter(
-                    X_valid_baseline,
+                    X_all_baseline,
                     baseline_mean_fr,
                     color=baseline_color,
                     marker=marker_list[baseline_i],
                     s=marker_size,
                 )
                 ax_fr.fill_between(
-                    X_valid_baseline,
+                    X_all_baseline,
                     baseline_mean_fr - baseline_ste_fr,
                     baseline_mean_fr + baseline_ste_fr,
                     color=baseline_color,
                     alpha=0.5,
                 )
 
+            # Seldonian failure rate
             for seldonian_i, seldonian_model in enumerate(seldonian_models):
                 this_seldonian_dict = seldonian_dict[seldonian_model]
                 seldonian_color = plot_colormap(seldonian_i)
                 df_seldonian = this_seldonian_dict["df_seldonian"]
                 n_trials = df_seldonian["trial_i"].max() + 1
-                mean_fr = df_seldonian.groupby("data_frac").mean()["failed"]
-                std_fr = df_seldonian.groupby("data_frac").std()["failed"]
+
+                gstr_failed = 'g' + str(constraint_num) + '_failed'
+
+                mean_fr = df_seldonian.groupby("data_frac").mean()[
+                    gstr_failed].to_numpy()
+                # Need to groupby data frac
+                std_fr = df_seldonian.groupby("data_frac").std()[
+                    gstr_failed].to_numpy()
+
                 ste_fr = std_fr / np.sqrt(n_trials)
 
-                X_all_seldonian = this_seldonian_dict["X_all"]
-                
+                X_all_seldonian = this_seldonian_dict["X_all"].to_numpy()
+
                 ax_fr.plot(
                     X_all_seldonian,
                     mean_fr,
@@ -549,8 +573,7 @@ class PlotGenerator:
             )
 
         if savename:
-            # plt.savefig(savename,format=save_format,dpi=600)
-            plt.savefig(savename, format=save_format)
+            plt.savefig(savename, format=save_format,bbox_inches="tight")
             print(f"Saved {savename}")
         else:
             plt.show()
@@ -678,7 +701,6 @@ class SupervisedPlotGenerator(PlotGenerator):
         sd_exp.run_experiment(**run_seldonian_kwargs)
         return
 
-
     def run_headless_seldonian_experiment(
         self, 
         full_pretraining_model,
@@ -745,16 +767,10 @@ class SupervisedPlotGenerator(PlotGenerator):
         sd_exp.run_experiment(**run_kwargs)
         return
 
-
-
-    def run_baseline_experiment(self, model_name, verbose=False):
+    def run_baseline_experiment(self, baseline_model, verbose=False):
         """Run a supervised Seldonian experiment using the spec attribute
         assigned to the class in __init__().
 
-        :param model_name: The name of the baseline model to use
-
-        :type model_name: str
-
         :param verbose: Whether to display results to stdout
                 while the Seldonian algorithms are running in each trial
         :type verbose: bool, defaults to False
@@ -785,7 +801,7 @@ class SupervisedPlotGenerator(PlotGenerator):
         )
 
         ## Run experiment
-        bl_exp = BaselineExperiment(model_name=model_name, results_dir=self.results_dir)
+        bl_exp = BaselineExperiment(baseline_model=baseline_model, results_dir=self.results_dir)
 
         bl_exp.run_experiment(**run_baseline_kwargs)
         return
